// Authors: Korbinian Schneeberger and Joerg Hagmann
// Copyright (C) 2008 by Max-Planck Institute for Developmental Biology, Tuebingen, Germany

#include "gmindex.h"

int init_defaults();
int init_opts(int argc, char *argv[]);
int init_index();
int init_constants();
int init_genome_file();
int init_index_file();
int init_meta_index_file();
int init_strain_file();
int init_genome_out_file();
//int init_stat_file();
int init_mem_master();
int init_sequence_container();

int init(int argc, char *argv[]) 
{
	init_defaults();
	init_opts(argc, argv);
	//init_index();
	init_constants();
	init_genome_file();
	init_index_file();
	init_strain_file();
	//if (strlen(OCC_FILE_NAME) > 0) init_stat_file();
	if (strlen(GENOME_OUT_FILE_NAME) > 0) init_genome_out_file();
	init_meta_index_file();
	init_mem_master();
	//init_sequence_container();
	
	//init_slot_variables();
	SLOT 	 = (unsigned int *) malloc(NUM_STRAINS * sizeof(unsigned int));
	SLOT_REV = (unsigned int *) malloc(NUM_STRAINS * sizeof(unsigned int));
	
	alloc_blocklist();
	alloc_blocktable();
	alloc_branchseqs();

	return 0;
}

int init_defaults() 
{
	DEBUG = 0;

	int i;

	NUM_USED_SLOTS = 0;
	
	VERBOSE = 0;
	SLOT_COUNTER = 0;
	POSITION_COUNTER = 0;
	INDEX_DEPTH = 12;
	LONGEST_CHROMOSOME = 0;
	BUILD_REVERSE_INDEX = 1;

	POWER[0] = 1;
	for(i = 1 ; i <= MAX_INDEX_DEPTH; i++) {
		POWER[i] = POWER[i-1] * 4;
	}

	return 0;
}

int init_opts(int argc, char *argv[]) 
{
	int i;
	char has_genome = 0;
	char has_index = 0;
	char has_meta_index = 0;
	char has_strains = 0;

	for (i = 1; i < argc; i++) {

		//genome IN
		if(strcmp(argv[i],"-i")==0){
			if(i+1 > argc - 1){ usage(); exit(1); }
			i++;
			strcpy(GENOME_FILE_NAME, argv[i]);
			has_genome = 1;
		}
		
		//genome OUT
		/*if(strcmp(argv[i],"-go")==0){
			if(i+1 > argc - 1){ usage(); exit(1); }
			i++;
			strcpy(GENOME_OUT_FILE_NAME, argv[i]);
			has_genome++;
		}*/
	
		//index
		if(strcmp(argv[i],"-x")==0){
			if(i+1 > argc - 1){ usage(); exit(1); }
			i++;
			strcpy(INDEX_FILE_NAME, argv[i]);
			has_index = 1;
		}

		//metaindex
		if(strcmp(argv[i],"-t")==0){
			if(i+1 > argc - 1){ usage(); exit(1); }
			i++;
			strcpy(META_INDEX_FILE_NAME, argv[i]);
			has_meta_index = 1;
		}
		
		//other strain
		if(strcmp(argv[i],"-d")==0){
			if(i+1 > argc - 1){ usage(); exit(1); }
			i++;
			STRAIN_FILE_NAME[0] = strtok(argv[i], ",");
			int k = 1; 
			while ((STRAIN_FILE_NAME[k] = strtok('\0', ",")) != NULL) k++;
		
			NUM_STRAINS = k;
			
			has_strains = 1;
		}
		
/*		//blocktable out
		if(strcmp(argv[i],"-b")==0){
			if(i+1 > argc - 1){ usage(); exit(1); }
			i++;
			char bla[100];
			strcpy(bla, argv[i]);
			BLOCKOUT = fopen(bla, "w");
		}
	*/	
		// seed occurrence stat output
/*		if(strcmp(argv[i],"-o")==0){
			if(i+1 > argc - 1){ usage(); exit(1); }
			i++;
			strcpy(OCC_FILE_NAME, argv[i]);
			has_meta_index = 1;
		}*/

		//depth
		if(strcmp(argv[i],"-s")==0){
			if(i+1 > argc - 1){ usage(); exit(1); }
			i++;
			if ((INDEX_DEPTH = atoi(argv[i])) == 0) {
				fprintf(stderr, "ERROR: seedlength must be an integer value between 5 and 12 (both inclusive)!\n");
				exit(0);
			}
			if (INDEX_DEPTH < MIN_INDEX_DEPTH) {
				fprintf(stderr, "ERROR: specified index depth too low. Minimum allowed is 5!\n");
				exit(1);
			}
			else if (INDEX_DEPTH > MAX_INDEX_DEPTH) {
				fprintf(stderr, "ERROR: specified index depth too big. Maximum allowed is 12!\n");
				exit(1);
			} 
		}
	
		//verbose
		if(strcmp(argv[i],"-v")==0){
			VERBOSE = 1;
		}
		
		//build reverse index
		if(strcmp(argv[i],"-r")==0){
			BUILD_REVERSE_INDEX = 0;
		}

		//debug
		if(strcmp(argv[i],"-p")==0){
			DEBUG = 1;
		}

  	}

	if (has_index == 0 || has_genome == 0 || has_meta_index == 0) {
		usage(); exit(1);
	}
	
	if (has_strains == 0) NUM_STRAINS = 0;
	
	// just as a control because it's so crucial:
	if (BLOCK_SIZE < INDEX_DEPTH) {
		fprintf(stderr, "ERROR: Block size smaller than Index depth! This doesn't work out!\n");
		exit(0);
	}

	return 1;
}

/*
int init_index()
{
	INDEX_SIZE = 4^INDEX_DEPTH;
	
	INDEX = (BIN *) malloc(INDEX_SIZE * sizeof(BIN));
	INDEX_REV = (BIN *) malloc(INDEX_SIZE * sizeof(BIN));
	
	return 1;
}*/


int init_constants() 
{
	if (INDEX_DEPTH == 5) {
		BINARY_CODE[0] = 0;			//binary number: 0000 0000 0000 0000 0000 0000
		BINARY_CODE[1] = 256;		//binary number: 0000 0000 0000 0001 0000 0000
		BINARY_CODE[2] = 512;		//binary number: 0000 0000 0000 0010 0000 0000
		BINARY_CODE[3] = 768;		//binary number: 0000 0000 0000 0011 0000 0000
		BINARY_CODE[4] = 1023;		//binary number: 0000 0000 0000 0011 1111 1111
	}
	if (INDEX_DEPTH == 6) {
		BINARY_CODE[0] = 0;			//binary number: 0000 0000 0000 0000 0000 0000
		BINARY_CODE[1] = 1024;		//binary number: 0000 0000 0000 0100 0000 0000
		BINARY_CODE[2] = 2048;		//binary number: 0000 0000 0000 1000 0000 0000
		BINARY_CODE[3] = 3072;		//binary number: 0000 0000 0000 1100 0000 0000
		BINARY_CODE[4] = 4095;		//binary number: 0000 0000 0000 1111 1111 1111
	}
	if (INDEX_DEPTH == 7) {
		BINARY_CODE[0] = 0;			//binary number: 0000 0000 0000 0000 0000 0000
		BINARY_CODE[1] = 4096;		//binary number: 0000 0000 0001 0000 0000 0000
		BINARY_CODE[2] = 8192;		//binary number: 0000 0000 0010 0000 0000 0000
		BINARY_CODE[3] = 12288;		//binary number: 0000 0000 0011 0000 0000 0000
		BINARY_CODE[4] = 16383;		//binary number: 0000 0000 0011 1111 1111 1111
	}
	if (INDEX_DEPTH == 8) {
		BINARY_CODE[0] = 0;			//binary number: 0000 0000 0000 0000 0000 0000
		BINARY_CODE[1] = 16384;		//binary number: 0000 0000 0100 0000 0000 0000
		BINARY_CODE[2] = 32768;		//binary number: 0000 0000 1000 0000 0000 0000
		BINARY_CODE[3] = 49152;		//binary number: 0000 0000 1100 0000 0000 0000
		BINARY_CODE[4] = 65535;		//binary number: 0000 0000 1111 1111 1111 1111
	}
	if (INDEX_DEPTH == 9) {
		BINARY_CODE[0] = 0;			//binary number: 0000 0000 0000 0000 0000 0000
		BINARY_CODE[1] = 65536;		//binary number: 0000 0001 0000 0000 0000 0000
		BINARY_CODE[2] = 131072;	//binary number: 0000 0010 0000 0000 0000 0000
		BINARY_CODE[3] = 196608;	//binary number: 0000 0011 0000 0000 0000 0000
		BINARY_CODE[4] = 262143;	//binary number: 0000 0011 1111 1111 1111 1111
	}
	if (INDEX_DEPTH == 10) {
		BINARY_CODE[0] = 0;			//binary number: 0000 0000 0000 0000 0000 0000
		BINARY_CODE[1] = 262144;	//binary number: 0000 0100 0000 0000 0000 0000
		BINARY_CODE[2] = 524288;	//binary number: 0000 1000 0000 0000 0000 0000
		BINARY_CODE[3] = 786432;	//binary number: 0000 1100 0000 0000 0000 0000
		BINARY_CODE[4] = 1048575;	//binary number: 0000 1111 1111 1111 1111 1111
	}
	if (INDEX_DEPTH == 11) {
		BINARY_CODE[0] = 0;			//binary number: 0000 0000 0000 0000 0000 0000
		BINARY_CODE[1] = 1048576;	//binary number: 0001 0000 0000 0000 0000 0000
		BINARY_CODE[2] = 2097152;	//binary number: 0010 0000 0000 0000 0000 0000
		BINARY_CODE[3] = 3145728;	//binary number: 0011 0000 0000 0000 0000 0000
		BINARY_CODE[4] = 4194303;	//binary number: 0011 1111 1111 1111 1111 1111
	}
	if (INDEX_DEPTH == 12) {
		BINARY_CODE[0] = 0;			//binary number: 0000 0000 0000 0000 0000 0000
		BINARY_CODE[1] = 4194304;	//binary number: 0100 0000 0000 0000 0000 0000
		BINARY_CODE[2] = 8388608;	//binary number: 1000 0000 0000 0000 0000 0000
		BINARY_CODE[3] = 12582912;	//binary number: 1100 0000 0000 0000 0000 0000
		BINARY_CODE[4] = 16777215;	//binary number: 1111 1111 1111 1111 1111 1111
	}
	if (INDEX_DEPTH == 13) {
        BINARY_CODE[0] = 0;             //binary number: 0000 0000 0000 0000 0000 0000 0000
        BINARY_CODE[1] = 16777216;      //binary number: 0001 0000 0000 0000 0000 0000 0000
        BINARY_CODE[2] = 33554432;      //binary number: 0010 0000 0000 0000 0000 0000 0000
        BINARY_CODE[3] = 50331648;      //binary number: 0011 0000 0000 0000 0000 0000 0000
        BINARY_CODE[4] = 67108863;		//binary number: 0011 1111 1111 1111 1111 1111 1111
    }

	return(0);
}

int init_genome_file() 
{
	char line[513], header = 0;

	if ((GENOME_FP = fopen(GENOME_FILE_NAME, "r")) == NULL) {
		fprintf(stderr, "ERROR : Couldn't open genome file %s\n", GENOME_FILE_NAME);
		exit(1);
	}

	do {
		if (fgets(line, 512, GENOME_FP) == NULL) {
			fprintf(stderr, "ERROR : Couldn't read input file %s\n", GENOME_FILE_NAME);
			exit(1);
		}
	} while (line[0] != '>');
	
	if (strchr(line, '\n') == NULL) header = 1;

	int chrdesclen = (strcspn(line, " \t") > CHR_DESC_LENGTH-1)? CHR_DESC_LENGTH-1: strcspn(line, " \t") - 2;
	if (chrdesclen > 0) strncpy(CHR_DESC, &line[1], chrdesclen);
		else {
			fprintf(stderr, "ERROR: A chromosome doesn't have a valid description!\n");
			exit(0);
		}
	CHR_DESC[chrdesclen] = '\0';
	strcpy(CHR_DESC_TMP, CHR_DESC);
	
	if (header == 1) {
		do {
			if (fgets(line, 512, GENOME_FP) == NULL) {
				fprintf(stderr, "ERROR : input file %s corrupt?\n", GENOME_FILE_NAME);
				exit(1);
			}
			if (strchr(line, '\n') != NULL) header = 0;
		} while (header != 0);
	}
	
	return 0;
}


int init_index_file() 
{
	if ((INDEX_FP = fopen(INDEX_FILE_NAME, "w")) == NULL) {
		fprintf(stderr, "ERROR : Couldn't open index file %s\n", INDEX_FILE_NAME);
		exit(1);
	}

	return(0);
}

int init_meta_index_file() 
{
	if ((META_INDEX_FP = fopen(META_INDEX_FILE_NAME, "w")) == NULL) {
		fprintf(stderr, "ERROR : Couldn't open meta index file %s\n", META_INDEX_FILE_NAME);
		exit(1);
	}

	return(0);
}

int init_strain_file() 
{
	STRAIN_FP = (FILE **) malloc(NUM_STRAINS * sizeof(FILE **));
	int i;
	for (i=0; i!=NUM_STRAINS; ++i) {
		STRAIN_FP[i] = (FILE *) malloc(sizeof(FILE *));
		if ((STRAIN_FP[i] = fopen(STRAIN_FILE_NAME[i], "r")) == NULL) {
			fprintf(stderr, "ERROR : Couldn't open strain file %s\n", STRAIN_FILE_NAME[i]);
			exit(1);
		}
	}

	return(0);
}

int init_genome_out_file() 
{
	if ((GENOME_OUT_FP = fopen(GENOME_OUT_FILE_NAME, "w")) == NULL) {
		fprintf(stderr, "ERROR : Couldn't open meta index file %s\n", GENOME_OUT_FILE_NAME);
		exit(1);
	}

	return(0);
}

/*int init_stat_file()
{
	if ((OCC_FP = fopen(OCC_FILE_NAME, "w")) == NULL) {
		fprintf(stderr, "ERROR : Couldn't open stat file %s\n", OCC_FILE_NAME);
		exit(1);
	}

	return(0);
}*/

int init_mem_master() 
{
	if ( (MEM_MGR = (STORAGE *) malloc(sizeof(STORAGE))) == NULL) {
		fprintf(stderr, "ERROR : not enough memory for memory manager\n");
		exit(1);
	}

	MEM_MGR->curr_num = 0;
	MEM_MGR->nuggets = NULL;

	return 0;
}

/*
int init_sequence_container()
{
	SEQ_FIRST = alloc_seq_container();
	SEQS = SEQ_FIRST;
	
	return 0;
} */

